@@ -6,65 +6,76 @@ from typing import (Iterable,
                     List,
                     Optional,
                     Sequence,
-                    Tuple,
-                    Union as Union_)
+                    Union)
 
 from ground.base import Context
-from ground.hints import (Point,
-                          Segment)
+from ground.hints import (Empty,
+                          Linear,
+                          Maybe,
+                          Mix,
+                          Multipoint,
+                          Point,
+                          Shaped)
 from reprit.base import generate_repr
 
 from . import bounding
-from .event import (ShapedEvent as Event,
-                    event_to_segment_endpoints,
+from .event import (LeftHolelessEvent as LeftEvent,
+                    RightShapedEvent as RightEvent,
                     events_to_connectivity)
 from .events_queue import HolelessEventsQueue as EventsQueue
-from .hints import (HolelessMix,
-                    Multiregion,
+from .hints import (Region,
                     SegmentEndpoints)
+from .operands import HolelessOperand
 from .sweep_line import BinarySweepLine as SweepLine
+from .unpacking import (unpack_mix,
+                        unpack_points,
+                        unpack_regions,
+                        unpack_segments)
 from .utils import (all_equal,
                     contour_to_oriented_edges_endpoints,
                     endpoints_to_segments,
                     pairwise,
-                    to_multiregion_x_max)
+                    to_endpoints,
+                    to_regions_x_max)
+
+Event = Union[LeftEvent, RightEvent]
 
 
 class Operation(ABC):
-    __slots__ = 'context', 'left', 'right', '_events_queue'
+    __slots__ = 'context', 'first', 'second', '_events_queue'
 
     def __init__(self,
-                 left: Multiregion,
-                 right: Multiregion,
+                 first: HolelessOperand,
+                 second: HolelessOperand,
                  context: Context) -> None:
         """
         Initializes operation.
 
-        :param left: left operand.
-        :param right: right operand.
+        :param first: first operand.
+        :param second: second operand.
         :param context: operation context.
         """
-        self.context, self.left, self.right = context, left, right
+        self.context, self.first, self.second = context, first, second
         self._events_queue = EventsQueue(context)
 
     __repr__ = generate_repr(__init__)
 
     @abstractmethod
-    def compute(self) -> Union_[Multiregion, HolelessMix]:
+    def compute(self) -> Union[Empty, Linear, Mix, Multipoint, Shaped]:
         """
         Computes result of the operation.
         """
 
-    def compute_fields(self, event: Event, below_event: Optional[Event]
+    def compute_fields(self, event: LeftEvent, below_event: Optional[LeftEvent]
                        ) -> None:
         if below_event is not None:
             event.other_interior_to_left = (below_event.other_interior_to_left
-                                            if (event.from_left
-                                                is below_event.from_left)
+                                            if (event.from_first
+                                                is below_event.from_first)
                                             else below_event.interior_to_left)
         event.in_result = self.in_result(event)
 
-    def events_to_multiregion(self, events: Iterable[Event]) -> Multiregion:
+    def events_to_regions(self, events: Iterable[Event]) -> Sequence[Region]:
         events = sorted([event for event in events if event.primary.in_result],
                         key=self._events_queue.key)
         for index, event in enumerate(events):
@@ -80,46 +91,43 @@ class Operation(ABC):
             vertices = [contour_start]
             contour_events = [event]
             cursor = event
-            complement_position = event.complement.position
-            processed[index] = processed[complement_position] = True
+            opposite_position = event.opposite.position
+            processed[index] = processed[opposite_position] = True
             while cursor.end != contour_start:
                 vertices.append(cursor.end)
-                position = _to_next_position(complement_position, processed,
+                position = _to_next_position(opposite_position, processed,
                                              connectivity)
                 if position is None:
                     break
                 cursor = events[position]
                 contour_events.append(cursor)
-                complement_position = cursor.complement.position
-                processed[position] = processed[complement_position] = True
+                opposite_position = cursor.opposite.position
+                processed[position] = processed[opposite_position] = True
             result.append(contour_cls(vertices))
         return result
 
     def fill_queue(self) -> None:
         events_queue = self._events_queue
-        for region in self.left:
+        for region in self.first.regions:
             events_queue.register(
                     contour_to_oriented_edges_endpoints(region, self.context),
                     True)
-        for region in self.right:
+        for region in self.second.regions:
             events_queue.register(
                     contour_to_oriented_edges_endpoints(region, self.context),
                     False)
 
     @abstractmethod
-    def in_result(self, event: Event) -> bool:
+    def in_result(self, event: LeftEvent) -> bool:
         """Detects if event will be presented in result of the operation."""
 
-    def normalize_operands(self) -> None:
-        pass
-
     def process_event(self,
                       event: Event,
                       processed_events: List[Event],
                       sweep_line: SweepLine) -> None:
-        if event.is_right_endpoint:
+        if not event.is_left:
             processed_events.append(event)
-            event = event.complement
+            event = event.left
             if event in sweep_line:
                 above_event, below_event = (sweep_line.above(event),
                                             sweep_line.below(event))
@@ -149,43 +157,20 @@ class Operation(ABC):
 class CompleteIntersection(Operation):
     __slots__ = ()
 
-    def compute(self) -> HolelessMix:
-        points, segments, multiregion = self._compute()
-        return (self.context.multipoint_cls(points),
-                self.context.multisegment_cls(segments), multiregion)
-
-    def in_result(self, event: Event) -> bool:
-        return (event.inside
-                or not event.from_left and event.is_common_region_boundary)
-
-    def sweep(self) -> Iterable[Event]:
-        self.fill_queue()
-        result = []
-        sweep_line = SweepLine(self.context)
-        min_max_x = min(to_multiregion_x_max(self.left),
-                        to_multiregion_x_max(self.right))
-        while self._events_queue:
-            event = self._events_queue.pop()
-            if min_max_x < event.start.x:
-                break
-            self.process_event(event, result, sweep_line)
-        return result
-
-    def _compute(self) -> Tuple[Sequence[Point], Sequence[Segment],
-                                Multiregion]:
-        if not (self.left and self.right):
-            return [], [], []
-        left_box, right_box = (self.context.contours_box(self.left),
-                               self.context.contours_box(self.right))
-        if bounding.disjoint_with(left_box, right_box):
-            return [], [], []
-        self.left = bounding.to_intersecting_regions(right_box, self.left,
-                                                     self.context)
-        self.right = bounding.to_intersecting_regions(left_box, self.right,
-                                                      self.context)
-        if not (self.left and self.right):
-            return [], [], []
-        self.normalize_operands()
+    def compute(self) -> Union[Empty, Linear, Mix, Multipoint, Shaped]:
+        context = self.context
+        first_box, second_box = (context.contours_box(self.first.regions),
+                                 context.contours_box(self.second.regions))
+        if bounding.disjoint_with(first_box, second_box):
+            return context.empty
+        self.first.regions = bounding.to_intersecting_regions(
+                second_box, self.first.regions, context)
+        if not self.first.regions:
+            return context.empty
+        self.second.regions = bounding.to_intersecting_regions(
+                first_box, self.second.regions, context)
+        if not self.second.regions:
+            return context.empty
         events = sorted(self.sweep(),
                         key=self._events_queue.key)
         points = []  # type: List[Point]
@@ -193,53 +178,38 @@ class CompleteIntersection(Operation):
         for start, same_start_events in groupby(events,
                                                 key=attrgetter('start')):
             same_start_events = list(same_start_events)
-            if (all(event.is_right_endpoint or not event.in_result
-                    for event in same_start_events)
-                    and not all_equal(event.from_left
-                                      for event in same_start_events)):
+            if not (any(event.is_left and event.in_result
+                        for event in same_start_events)
+                    or all_equal(event.from_first
+                                 for event in same_start_events)):
                 no_segment_found = True
                 for event, next_event in pairwise(same_start_events):
-                    if (event.from_left is not next_event.from_left
+                    if (event.from_first is not next_event.from_first
                             and event.start == next_event.start
                             and event.end == next_event.end):
                         no_segment_found = False
-                        if not event.is_right_endpoint:
-                            endpoints.append(
-                                    event_to_segment_endpoints(next_event))
-                if no_segment_found and all(not event.complement.in_result
-                                            if event.is_right_endpoint
-                                            else not event.in_result
+                        if event.is_left:
+                            endpoints.append(to_endpoints(next_event))
+                if no_segment_found and all(not event.primary.in_result
                                             for event in same_start_events):
                     points.append(start)
-        return (points, endpoints_to_segments(endpoints, self.context),
-                self.events_to_multiregion(events))
+        segments = endpoints_to_segments(endpoints, context)
+        regions = self.events_to_regions(events)
+        return unpack_mix(unpack_points(points, context),
+                          unpack_segments(segments, context),
+                          unpack_regions(regions, context),
+                          context)
 
-
-class Intersection(Operation):
-    __slots__ = ()
-
-    def compute(self) -> Multiregion:
-        if not (self.left and self.right):
-            return []
-        left_box, right_box = (self.context.contours_box(self.left),
-                               self.context.contours_box(self.right))
-        if bounding.disjoint_with(left_box, right_box):
-            return []
-        self.left = bounding.to_coupled_regions(right_box, self.left,
-                                                self.context)
-        self.right = bounding.to_coupled_regions(left_box, self.right,
-                                                 self.context)
-        if not (self.left and self.right):
-            return []
-        self.normalize_operands()
-        return self.events_to_multiregion(self.sweep())
+    def in_result(self, event: LeftEvent) -> bool:
+        return (event.inside
+                or not event.from_first and event.is_common_region_boundary)
 
     def sweep(self) -> Iterable[Event]:
         self.fill_queue()
         result = []
         sweep_line = SweepLine(self.context)
-        min_max_x = min(to_multiregion_x_max(self.left),
-                        to_multiregion_x_max(self.right))
+        min_max_x = min(to_regions_x_max(self.first.regions),
+                        to_regions_x_max(self.second.regions))
         while self._events_queue:
             event = self._events_queue.pop()
             if min_max_x < event.start.x:
@@ -247,9 +217,42 @@ class Intersection(Operation):
             self.process_event(event, result, sweep_line)
         return result
 
-    def in_result(self, event: Event) -> bool:
+
+class Intersection(Operation):
+    __slots__ = ()
+
+    def compute(self) -> Maybe[Shaped]:
+        context = self.context
+        first_box, second_box = (context.contours_box(self.first.regions),
+                                 context.contours_box(self.second.regions))
+        if bounding.disjoint_with(first_box, second_box):
+            return context.empty
+        self.first.regions = bounding.to_coupled_regions(
+                second_box, self.first.regions, context)
+        if not self.first.regions:
+            return context.empty
+        self.second.regions = bounding.to_coupled_regions(
+                first_box, self.second.regions, context)
+        return (unpack_regions(self.events_to_regions(self.sweep()), context)
+                if self.second.regions
+                else context.empty)
+
+    def sweep(self) -> Iterable[Event]:
+        self.fill_queue()
+        result = []
+        events_queue, sweep_line = self._events_queue, SweepLine(self.context)
+        min_max_x = min(to_regions_x_max(self.first.regions),
+                        to_regions_x_max(self.second.regions))
+        while events_queue:
+            event = events_queue.pop()
+            if min_max_x < event.start.x:
+                break
+            self.process_event(event, result, sweep_line)
+        return result
+
+    def in_result(self, event: LeftEvent) -> bool:
         return (event.inside
-                or not event.from_left and event.is_common_region_boundary)
+                or not event.from_first and event.is_common_region_boundary)
 
 
 def _to_next_position(position: int,
