@@ -6,48 +6,59 @@ from typing import (Iterable,
                     List,
                     Optional,
                     Sequence,
-                    Tuple,
                     Union as Union_)
 
 from ground.base import Context
-from ground.hints import (Multipolygon,
+from ground.hints import (Contour,
+                          Empty,
+                          Mix,
+                          Multipoint,
+                          Multipolygon,
+                          Multisegment,
                           Point,
                           Polygon,
                           Segment)
 from reprit.base import generate_repr
 
 from . import bounding
-from .event import (HoleyEvent as Event,
-                    event_to_segment_endpoints,
+from .event import (LeftHoleyEvent as LeftEvent,
+                    RightShapedEvent as RightEvent,
                     events_to_connectivity)
 from .events_queue import HoleyEventsQueue as EventsQueue
-from .hints import (Mix,
-                    SegmentEndpoints)
+from .hints import SegmentEndpoints
+from .operands import HoleyOperand
 from .sweep_line import BinarySweepLine as SweepLine
+from .unpacking import (unpack_mix,
+                        unpack_points,
+                        unpack_polygons,
+                        unpack_segments)
 from .utils import (all_equal,
                     endpoints_to_segments,
                     pairwise,
                     polygon_to_oriented_edges_endpoints,
                     shrink_collinear_vertices,
+                    to_endpoints,
                     to_first_border_vertex,
                     to_polygons_x_max)
 
+Event = Union_[LeftEvent, RightEvent]
+
 
 class Operation(ABC):
-    __slots__ = 'context', 'left', 'right', '_events_queue'
+    __slots__ = 'context', 'first', 'second', '_events_queue'
 
     def __init__(self,
-                 left: Sequence[Polygon],
-                 right: Sequence[Polygon],
+                 first: HoleyOperand,
+                 second: HoleyOperand,
                  context: Context) -> None:
         """
         Initializes operation.
 
-        :param left: left operand.
-        :param right: right operand.
+        :param first: first operand.
+        :param second: second operand.
         :param context: operation context.
         """
-        self.context, self.left, self.right = context, left, right
+        self.context, self.first, self.second = context, first, second
         self._events_queue = EventsQueue(context)
 
     __repr__ = generate_repr(__init__)
@@ -58,12 +69,12 @@ class Operation(ABC):
         Computes result of the operation.
         """
 
-    def compute_fields(self, event: Event, below_event: Optional[Event]
+    def compute_fields(self, event: LeftEvent, below_event: Optional[LeftEvent]
                        ) -> None:
         if below_event is not None:
             event.other_interior_to_left = (below_event.other_interior_to_left
-                                            if (event.from_left
-                                                is below_event.from_left)
+                                            if (event.from_first
+                                                is below_event.from_first)
                                             else below_event.interior_to_left)
             event.below_in_result_event = (below_event.below_in_result_event
                                            if (not self.in_result(below_event)
@@ -72,14 +83,15 @@ class Operation(ABC):
         event.in_result = self.in_result(event)
 
     def events_to_polygons(self, events: Iterable[Event]) -> Sequence[Polygon]:
-        events = sorted([event for event in events if event.primary.in_result],
-                        key=self._events_queue.key)
+        events = [event for event in events if event.primary.in_result]
+        events.sort(key=self._events_queue.key)
         for index, event in enumerate(events):
             event.position = index
         are_internal, depths, holes, parents = [], [], [], []
         processed = [False] * len(events)
-        contour_cls = self.context.contour_cls
-        contours = []
+        context = self.context
+        contour_cls = context.contour_cls
+        contours = []  # type: List[Contour]
         connectivity = events_to_connectivity(events)
         for index, event in enumerate(events):
             if processed[index]:
@@ -89,13 +101,13 @@ class Operation(ABC):
                                parents)
             vertices = _events_to_contour_vertices(event, events, contour_id,
                                                    connectivity, processed)
-            shrink_collinear_vertices(vertices, self.context)
+            shrink_collinear_vertices(vertices, context)
             if depths[contour_id] % 2:
                 # holes will be in clockwise order
                 vertices[:] = vertices[:1] + vertices[:0:-1]
             contours.append(contour_cls(vertices))
         result = []
-        polygon_cls = self.context.polygon_cls
+        polygon_cls = context.polygon_cls
         for index, contour in enumerate(contours):
             if are_internal[index]:
                 # hole of a hole is an external polygon
@@ -112,29 +124,26 @@ class Operation(ABC):
 
     def fill_queue(self) -> None:
         events_queue = self._events_queue
-        for polygon in self.left:
+        for polygon in self.first.polygons:
             events_queue.register(
                     polygon_to_oriented_edges_endpoints(polygon, self.context),
                     True)
-        for polygon in self.right:
+        for polygon in self.second.polygons:
             events_queue.register(
                     polygon_to_oriented_edges_endpoints(polygon, self.context),
                     False)
 
     @abstractmethod
-    def in_result(self, event: Event) -> bool:
+    def in_result(self, event: LeftEvent) -> bool:
         """Detects if event will be presented in result of the operation."""
 
-    def normalize_operands(self) -> None:
-        pass
-
     def process_event(self,
                       event: Event,
                       processed_events: List[Event],
                       sweep_line: SweepLine) -> None:
-        if event.is_right_endpoint:
+        if not event.is_left:
             processed_events.append(event)
-            event = event.complement
+            event = event.left
             if event in sweep_line:
                 above_event, below_event = (sweep_line.above(event),
                                             sweep_line.below(event))
@@ -173,12 +182,21 @@ class Operation(ABC):
 class Difference(Operation):
     __slots__ = ()
 
-    def compute(self) -> Multipolygon:
-        return self.context.multipolygon_cls(self._compute())
+    def compute(self) -> Union_[Empty, Multipolygon, Polygon]:
+        context = self.context
+        first_box = context.polygons_box(self.first.polygons)
+        if bounding.disjoint_with(first_box,
+                                  context.polygons_box(self.second.polygons)):
+            return self.first.value
+        self.second.polygons = bounding.to_coupled_polygons(
+                first_box, self.second.polygons, context)
+        return (unpack_polygons(self.events_to_polygons(self.sweep()), context)
+                if self.second.polygons
+                else self.first.value)
 
-    def in_result(self, event: Event) -> bool:
+    def in_result(self, event: LeftEvent) -> bool:
         return (event.outside
-                if event.from_left
+                if event.from_first
                 else event.inside or event.is_common_polyline_component)
 
     def sweep(self) -> Iterable[Event]:
@@ -186,72 +204,33 @@ class Difference(Operation):
         result = []
         events_queue = self._events_queue
         sweep_line = SweepLine(self.context)
-        left_x_max = to_polygons_x_max(self.left)
+        first_x_max = to_polygons_x_max(self.first.polygons)
         while events_queue:
             event = events_queue.pop()
-            if left_x_max < event.start.x:
+            if first_x_max < event.start.x:
                 break
             self.process_event(event, result, sweep_line)
         return result
 
-    def _compute(self) -> Sequence[Polygon]:
-        if not (self.left and self.right):
-            return self.left
-        left_box = self.context.polygons_box(self.left)
-        if bounding.disjoint_with(left_box,
-                                  self.context.polygons_box(self.right)):
-            return self.left
-        self.right = bounding.to_coupled_polygons(left_box, self.right,
-                                                  self.context)
-        if not self.right:
-            return self.left
-        self.normalize_operands()
-        return self.events_to_polygons(self.sweep())
-
 
 class CompleteIntersection(Operation):
     __slots__ = ()
 
-    def compute(self) -> Mix:
-        points, segments, polygons = self._compute()
+    def compute(self) -> Union_[Empty, Mix, Multipoint, Multipolygon,
+                                Multisegment, Polygon, Segment]:
         context = self.context
-        return (context.multipoint_cls(points),
-                context.multisegment_cls(segments),
-                context.multipolygon_cls(polygons))
-
-    def in_result(self, event: Event) -> bool:
-        return (event.inside
-                or not event.from_left and event.is_common_region_boundary)
-
-    def sweep(self) -> Iterable[Event]:
-        self.fill_queue()
-        result = []
-        events_queue = self._events_queue
-        sweep_line = SweepLine(self.context)
-        min_max_x = min(to_polygons_x_max(self.left),
-                        to_polygons_x_max(self.right))
-        while events_queue:
-            event = events_queue.pop()
-            if min_max_x < event.start.x:
-                break
-            self.process_event(event, result, sweep_line)
-        return result
-
-    def _compute(self) -> Tuple[Sequence[Point], Sequence[Segment],
-                                Sequence[Polygon]]:
-        if not (self.left and self.right):
-            return [], [], []
-        left_box, right_box = (self.context.polygons_box(self.left),
-                               self.context.polygons_box(self.right))
-        if bounding.disjoint_with(left_box, right_box):
-            return [], [], []
-        self.left = bounding.to_intersecting_polygons(right_box, self.left,
-                                                      self.context)
-        self.right = bounding.to_intersecting_polygons(left_box, self.right,
-                                                       self.context)
-        if not (self.left and self.right):
-            return [], [], []
-        self.normalize_operands()
+        first_box, second_box = (context.polygons_box(self.first.polygons),
+                                 context.polygons_box(self.second.polygons))
+        if bounding.disjoint_with(first_box, second_box):
+            return context.empty
+        self.first.polygons = bounding.to_intersecting_polygons(
+                second_box, self.first.polygons, context)
+        if not self.first.polygons:
+            return context.empty
+        self.second.polygons = bounding.to_intersecting_polygons(
+                first_box, self.second.polygons, context)
+        if not self.second.polygons:
+            return context.empty
         events = sorted(self.sweep(),
                         key=self._events_queue.key)
         points = []  # type: List[Point]
@@ -259,43 +238,39 @@ class CompleteIntersection(Operation):
         for start, same_start_events in groupby(events,
                                                 key=attrgetter('start')):
             same_start_events = list(same_start_events)
-            if (all(event.is_right_endpoint or not event.in_result
-                    for event in same_start_events)
-                    and not all_equal(event.from_left
-                                      for event in same_start_events)):
+            if not (any(event.is_left and event.in_result
+                        for event in same_start_events)
+                    or all_equal(event.from_first
+                                 for event in same_start_events)):
                 no_segment_found = True
                 for event, next_event in pairwise(same_start_events):
-                    if (event.from_left is not next_event.from_left
+                    if (event.from_first is not next_event.from_first
                             and event.start == next_event.start
                             and event.end == next_event.end):
                         no_segment_found = False
-                        if not event.is_right_endpoint:
-                            endpoints.append(
-                                    event_to_segment_endpoints(next_event))
+                        if event.is_left:
+                            endpoints.append(to_endpoints(next_event))
                 if no_segment_found and all(not event.primary.in_result
                                             for event in same_start_events):
                     points.append(start)
-        return (points, endpoints_to_segments(endpoints, self.context),
-                self.events_to_polygons(events))
+        segments = endpoints_to_segments(endpoints, context)
+        polygons = self.events_to_polygons(events)
+        return unpack_mix(unpack_points(points, context),
+                          unpack_segments(segments, context),
+                          unpack_polygons(polygons, context),
+                          context)
 
-
-class Intersection(Operation):
-    __slots__ = ()
-
-    def compute(self) -> Multipolygon:
-        return self.context.multipolygon_cls(self._compute())
-
-    def in_result(self, event: Event) -> bool:
+    def in_result(self, event: LeftEvent) -> bool:
         return (event.inside
-                or not event.from_left and event.is_common_region_boundary)
+                or not event.from_first and event.is_common_region_boundary)
 
     def sweep(self) -> Iterable[Event]:
         self.fill_queue()
         result = []
         events_queue = self._events_queue
         sweep_line = SweepLine(self.context)
-        min_max_x = min(to_polygons_x_max(self.left),
-                        to_polygons_x_max(self.right))
+        min_max_x = min(to_polygons_x_max(self.first.polygons),
+                        to_polygons_x_max(self.second.polygons))
         while events_queue:
             event = events_queue.pop()
             if min_max_x < event.start.x:
@@ -303,71 +278,83 @@ class Intersection(Operation):
             self.process_event(event, result, sweep_line)
         return result
 
-    def _compute(self) -> Sequence[Polygon]:
-        if not (self.left and self.right):
-            return []
-        left_box, right_box = (self.context.polygons_box(self.left),
-                               self.context.polygons_box(self.right))
-        if bounding.disjoint_with(left_box, right_box):
-            return []
-        self.left = bounding.to_coupled_polygons(right_box, self.left,
-                                                 self.context)
-        self.right = bounding.to_coupled_polygons(left_box, self.right,
-                                                  self.context)
-        if not (self.left and self.right):
-            return []
-        self.normalize_operands()
-        return self.events_to_polygons(self.sweep())
+
+class Intersection(Operation):
+    __slots__ = ()
+
+    def compute(self) -> Union_[Empty, Multipolygon, Polygon]:
+        context = self.context
+        first_box, second_box = (context.polygons_box(self.first.polygons),
+                                 context.polygons_box(self.second.polygons))
+        if bounding.disjoint_with(first_box, second_box):
+            return context.empty
+        self.first.polygons = bounding.to_coupled_polygons(
+                second_box, self.first.polygons, context)
+        if not self.first.polygons:
+            return context.empty
+        self.second.polygons = bounding.to_coupled_polygons(
+                first_box, self.second.polygons, context)
+        return (unpack_polygons(self.events_to_polygons(self.sweep()), context)
+                if self.second.polygons
+                else context.empty)
+
+    def in_result(self, event: LeftEvent) -> bool:
+        return (event.inside
+                or not event.from_first and event.is_common_region_boundary)
+
+    def sweep(self) -> Iterable[Event]:
+        self.fill_queue()
+        result = []
+        events_queue = self._events_queue
+        sweep_line = SweepLine(self.context)
+        min_max_x = min(to_polygons_x_max(self.first.polygons),
+                        to_polygons_x_max(self.second.polygons))
+        while events_queue:
+            event = events_queue.pop()
+            if min_max_x < event.start.x:
+                break
+            self.process_event(event, result, sweep_line)
+        return result
 
 
 class SymmetricDifference(Operation):
     __slots__ = ()
 
-    def compute(self) -> Multipolygon:
-        return self.context.multipolygon_cls(self._compute())
+    def compute(self) -> Union_[Empty, Multipolygon, Polygon]:
+        context = self.context
+        if bounding.disjoint_with(context.polygons_box(self.first.polygons),
+                                  context.polygons_box(self.second.polygons)):
+            polygons = []
+            polygons += self.first.polygons
+            polygons += self.second.polygons
+            polygons.sort(key=to_first_border_vertex)
+            return context.multipolygon_cls(polygons)
+        return unpack_polygons(self.events_to_polygons(self.sweep()), context)
 
-    def in_result(self, event: Event) -> bool:
+    def in_result(self, event: LeftEvent) -> bool:
         return not event.is_overlap
 
-    def _compute(self) -> Sequence[Polygon]:
-        if not (self.left and self.right):
-            return self.left or self.right
-        elif bounding.disjoint_with(self.context.polygons_box(self.left),
-                                    self.context.polygons_box(self.right)):
-            result = []
-            result += self.left
-            result += self.right
-            result.sort(key=to_first_border_vertex)
-            return result
-        self.normalize_operands()
-        return self.events_to_polygons(self.sweep())
-
 
 class Union(Operation):
     __slots__ = ()
 
-    def compute(self) -> Multipolygon:
-        return self.context.multipolygon_cls(self._compute())
+    def compute(self) -> Union_[Multipolygon, Polygon]:
+        context = self.context
+        if bounding.disjoint_with(context.polygons_box(self.first.polygons),
+                                  context.polygons_box(self.second.polygons)):
+            polygons = []
+            polygons += self.first.polygons
+            polygons += self.second.polygons
+            polygons.sort(key=to_first_border_vertex)
+            return context.multipolygon_cls(polygons)
+        return unpack_polygons(self.events_to_polygons(self.sweep()), context)
 
-    def _compute(self) -> Sequence[Polygon]:
-        if not (self.left and self.right):
-            return self.left or self.right
-        elif bounding.disjoint_with(self.context.polygons_box(self.left),
-                                    self.context.polygons_box(self.right)):
-            result = []
-            result += self.left
-            result += self.right
-            result.sort(key=to_first_border_vertex)
-            return result
-        self.normalize_operands()
-        return self.events_to_polygons(self.sweep())
-
-    def in_result(self, event: Event) -> bool:
+    def in_result(self, event: LeftEvent) -> bool:
         return (event.outside
-                or not event.from_left and event.is_common_region_boundary)
+                or not event.from_first and event.is_common_region_boundary)
 
 
-def _compute_relations(event: Event,
+def _compute_relations(event: LeftEvent,
                        contour_id: int,
                        are_internal: List[bool],
                        depths: List[int],
@@ -397,42 +384,41 @@ def _compute_relations(event: Event,
     are_internal.append(is_internal)
 
 
-def _events_to_contour_vertices(cursor: Event,
-                                events: Sequence[Event],
+def _events_to_contour_vertices(cursor: LeftEvent,
+                                events: Sequence[LeftEvent],
                                 contour_id: int,
                                 connectivity: Sequence[int],
                                 processed: List[bool]) -> List[Point]:
     contour_start = cursor.start
-    contour = [contour_start]
-    contour_events = [cursor]
-    complement_position = cursor.complement.position
+    result, contour_events = [contour_start], [cursor]
+    opposite_position = cursor.right.position
     vertices_positions = {contour_start: 0}
     while cursor.end != contour_start:
         vertex = cursor.end
         if vertex in vertices_positions:
             # vertices loop found, i.e. contour has self-intersection
             previous_vertex_position = vertices_positions[vertex]
-            del contour[previous_vertex_position:]
+            del result[previous_vertex_position:]
             del contour_events[previous_vertex_position:]
         else:
-            vertices_positions[vertex] = len(contour)
-        contour.append(vertex)
-        position = _to_next_position(complement_position, processed,
+            vertices_positions[vertex] = len(result)
+        result.append(vertex)
+        position = _to_next_position(opposite_position, processed,
                                      connectivity)
         if position is None:
             break
         cursor = events[position]
         contour_events.append(cursor)
-        complement_position = cursor.complement.position
+        opposite_position = cursor.opposite.position
     for event in contour_events:
-        processed[event.position] = processed[event.complement.position] = True
-        if event.is_right_endpoint:
-            event.complement.result_in_out = True
-            event.complement.contour_id = contour_id
-        else:
+        processed[event.position] = processed[event.opposite.position] = True
+        if event.is_left:
             event.result_in_out = False
             event.contour_id = contour_id
-    return contour
+        else:
+            event.opposite.result_in_out = True
+            event.opposite.contour_id = contour_id
+    return result
 
 
 def _to_next_position(position: int,
